#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>

#include "vector.h"
#include "Block.h"
#include "Piece.h"

struct _Piece {
  PieceType_t type;

  int count;

  vector_t rotate_angle;
  vector_t rotate_v;

  vector_t position;
  vector_t v;
  vector_t a;

  Block block;
};

Piece Piece_Create(Block block, PieceType_t type)
{
  Piece piece;

  piece = malloc(sizeof(*piece));
  memset(piece, 0, sizeof(*piece));

  piece->type = type;

  piece->count = rand() % 40 + 10;

  vector_set(&piece->rotate_angle, rand() % 360, rand() % 360, rand() % 360);
  vector_set(&piece->rotate_v,
	     (rand() % 61) - 30, (rand() % 61) - 30, (rand() % 61) - 30);

  vector_zero(&piece->position);
  vector_set(&piece->v,
	     ((rand() % 401) - 200) * 0.0001,
	     ((rand() % 401) - 200) * 0.0001,
	     ((rand() % 201) + 100) * 0.0001);
  vector_set(&piece->a, 0.0, 0.0, -0.001);

  piece->block = block;

  return piece;
}

Piece Piece_Destroy(Piece piece)
{
  if (piece) free(piece);
  return NULL;
}

int Piece_Proc(Piece piece)
{
  if (piece->count == 0)
    return 0;
  piece->count--;

  vector_add(&piece->rotate_angle, &piece->rotate_angle, &piece->rotate_v);
  vector_add(&piece->v, &piece->v, &piece->a);

  piece->v.x *= 0.95;
  piece->v.y *= 0.95;

  vector_add(&piece->position, &piece->position, &piece->v);

  return 1;
}

int Piece_MakePolygon(Piece piece)
{
  GLfloat red[]     = { 0.7, 0.0, 0.0, 0.6 };
  GLfloat magenta[] = { 0.7, 0.0, 0.7, 0.6 };
  GLfloat *color;

  switch (piece->type) {
  default:
  case PIECE_TYPE_FAIL:    color = magenta; break;
  case PIECE_TYPE_SUCCEED: color = red;     break;
  }

  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT , color);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE , color);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, color);

  glTranslated(piece->position.x, piece->position.y, piece->position.z);
  glRotated(piece->rotate_angle.x, 1.0, 0.0, 0.0);
  glRotated(piece->rotate_angle.y, 0.0, 1.0, 0.0);
  glRotated(piece->rotate_angle.z, 0.0, 0.0, 1.0);
  glScaled(0.02, 0.02, 0.02);

  glBegin(GL_POLYGON);
  glNormal3d( 0.0, 0.0, 1.0);
  glVertex3d( 1.0, 1.0, 0.0);
  glVertex3d(-1.0, 1.0, 0.0);
  glVertex3d(-1.0,-1.0, 0.0);
  glVertex3d( 1.0,-1.0, 0.0);
  glEnd();

  return 0;
}
