#!/usr/bin/perl

use Socket;
use IO::Handle;

################################
# Original code for Reporter.

$dir = ".";
$port = 12345;
$quiet = 0;
$omit = 0;
$replay = 0;
$speed = 1.0;

while (($_ = shift(@ARGV)) ne "") {
	if    ($_ eq "-dir"   ) { $dir    = shift(@ARGV); }
	elsif ($_ eq "-d"     ) { $dir    = shift(@ARGV); }
	elsif ($_ eq "-port"  ) { $port   = shift(@ARGV); }
	elsif ($_ eq "-p"     ) { $port   = shift(@ARGV); }
	elsif ($_ eq "-host"  ) { $host   = shift(@ARGV); }
	elsif ($_ eq "-h"     ) { $host   = shift(@ARGV); }
	elsif ($_ eq "-quiet" ) { $quiet  = 1;            }
	elsif ($_ eq "-q"     ) { $quiet  = 1;            }
	elsif ($_ eq "-omit"  ) { $omit   = 1;            }
	elsif ($_ eq "-o"     ) { $omit   = 1;            }
	elsif ($_ eq "-replay") { $replay = 1;            }
	elsif ($_ eq "-r"     ) { $replay = 1;            }
	elsif ($_ eq "-speed" ) { $speed  = shift(@ARGV); }
	elsif ($_ eq "-s"     ) { $speed  = shift(@ARGV); }
}

$addr = pack_sockaddr_in($port, INADDR_ANY);

if (!defined($host)) {
	socket(SOCK, PF_INET, SOCK_STREAM, 0) || die "socket()";
	setsockopt(SOCK, SOL_SOCKET, SO_REUSEADDR, 1);
	bind(SOCK, $addr) || die "bind()";
	listen(SOCK, SOMAXCONN) || die "listen()";
} else {
	socket(SOCKET, PF_INET, SOCK_STREAM, 0) || die "socket()";
	connect(SOCKET, $addr) || die "connect()";
	print_message("Connect.");
	SOCKET->autoflush;
	mainloop();
	exit(0);
}

while (1) {
	accept(SOCKET, SOCK) || die "accept()";
	print_message("Connected.");

	if (fork() == 0) { # Child process
		SOCKET->autoflush;
		close(SOCK);
		mainloop();
	}
	close(SOCKET);
}
exit(0);

sub mainloop {
	%uid_list = ();
	%qid_list = ();
	%reply_list = ();
	%clear_list = ();

	while (1) {
		$rin = "";
		vec($rin, fileno(SOCKET), 1) = 1;
		($n, $lefttime) = select($rout=$rin, undef, undef, 0);
		if ($n > 0) {
			$size = read(SOCKET, $line, 10);
			if ($size == 0) {
				print_message("Disconnected.");
				exit(0);
			}
		}

		sleep(0.25);

		@list = check_user_list();
		foreach $uid (@list) {
			$name = user_get_name($uid);
			send_message("USER ${uid} ${name}");
			$uid_list{$uid} = 1;
		}

		sleep(0.25);

		@list = check_question_list();
		foreach $qid (@list) {
			$genre = question_get_genre($qid);
			$score = question_get_score($qid);
			send_message("QUESTION ${qid} ${genre} ${score}");
			$qid_list{$qid} = 1;
		}

		if ($replay && !defined($toptime)) {
			@list = check_reply_list();
			foreach $reply (@list) {
				($uid, $qid, $t) = split('/', $reply);
				if (!defined($toptime) || ($t < $toptime)) {
					$toptime = $t;
				}
			}
			@list = check_clear_list();
			foreach $clear (@list) {
				($uid, $qid, $t) = split('/', $clear);
				if (!defined($toptime) || ($t < $toptime)) {
					$toptime = $t;
				}
			}
			if (!defined($toptime)) {
				next;
			}
			$starttime = time();
		}

		$nowtime = time();
		$difftime = ($nowtime - $starttime) * $speed;

		sleep(0.25);

		@list = check_reply_list();
		foreach $reply (@list) {
			($uid, $qid, $t) = split('/', $reply);
			if ($replay) {
				if ($difftime < $t - $toptime) {
					next;
				}
			}
			if (!$omit) {
				send_message("REPLY ${uid} ${qid} ${t}");
			}
			$reply_list{$reply} = 1;
		}

		sleep(0.25);

		@list = check_clear_list();
		foreach $clear (@list) {
			($uid, $qid, $t) = split('/', $clear);
			if ($replay) {
				if ($difftime < $t - $toptime) {
					next;
				}
			}
			send_message("CLEAR ${uid} ${qid} ${t}");
			$key = "${uid}/${qid}";
			$clear_list{$key} = $t;
		}
		if ($omit) {
			$omit = 0;
		}
	}
}

sub send_message {
	my $line;
	($line) = @_;
	print_message($line);
	print SOCKET "$line\n";
}

sub print_message {
	my $line;
	($line) = @_;
	if (!$quiet) {
		print "$line\n";
	}
}

sub check_user_list {
	my @list;
	my $uid;

	@list = ();
	foreach $uid (user_list()) {
		if (!defined($uid_list{$uid})) {
			push(@list, $uid);
		}
	}

	return @list;
}

sub check_question_list {
	my @list;
	my $qid;

	@list = ();
	foreach $qid (question_list()) {
		if (!defined($qid_list{$qid})) {
			push(@list, $qid);
		}
	}

	return @list;
}

sub check_reply_list {
	my $uid;
	my $qid;
	my @files;
	my $f;
	my @list;

	@list = ();
	foreach $qid (keys(%qid_list)) {
		foreach $uid (keys(%uid_list)) {
			@files = glob("$dir/result/${uid}/${qid}-*.txt");
			foreach $f (@files) {
				($t) = $f =~ /\-(\d+)\.txt$/;
				$key = "${uid}/${qid}/${t}";
				if (!defined($reply_list{$key})) {
					push(@list, $key);
				}
			}
		}
	}

	return @list;
}

sub check_clear_list {
	my $uid;
	my $qid;
	my @files;
	my $f;
	my @list;

	@list = ();
	foreach $qid (keys(%qid_list)) {
		foreach $uid (keys(%uid_list)) {
			$key = "${uid}/${qid}";
			if (!defined($clear_list{$key})) {
				if (question_is_clear($uid, $qid)) {
					$t = clear_get_time($uid, $qid);
					$key = "${uid}/${qid}/${t}";
					push(@list, $key);
				}
			}
		}
	}

	return @list;
}

################################
# From EasyCTF library.

sub user_list {
	my $uid;
	my $name;
	my $password;
	my @list;

	@list = ();

	open(PASSWORDFILE, "$dir/user/password.txt");
	while (<PASSWORDFILE>) {
		s/^\s+//;
		s/\s+$//;
		if (/^#/ || ($_ eq "")) {
			next;
		}
		($uid, $name, $password) = split('\s+');
		push(@list, $uid);
	}
	close(PASSWORDFILE);

	return @list;
}

sub user_get_uid {
	my $search_name;
	my $uid;
	my $name;
	my $password;

	($search_name) = @_;

	open(PASSWORDFILE, "$dir/user/password.txt");
	while (<PASSWORDFILE>) {
		s/^\s+//;
		s/\s+$//;
		if (/^#/ || ($_ eq "")) {
			next;
		}
		($uid, $name, $password) = split('\s+');
		if ($search_name ne $name) {
			next;
		}
		close(PASSWORDFILE);
		return $uid;
	}
	close(PASSWORDFILE);

	return "";
}

sub user_get_name {
	my $uid;
	my $name;
	my $password;

	($uid) = @_;
	($uid, $name, $password) = user_readfile($uid);
	return $name;
}

sub user_readfile {
	my $search_uid;
	my $uid;
	my $name;
	my $password;

	($search_uid) = @_;

	open(PASSWORDFILE, "$dir/user/password.txt");
	while (<PASSWORDFILE>) {
		s/^\s+//;
		s/\s+$//;
		if (/^#/ || ($_ eq "")) {
			next;
		}
		($uid, $name, $password) = split('\s+');
		if ($search_uid ne $uid) {
			next;
		}
		close(PASSWORDFILE);
		return ($uid, $name, $password);
	}
	close(PASSWORDFILE);

	return "";
}

sub question_list {
	my $qid;
	my $file;
	my @list;

	@list = ();

	open(QUESTIONLIST, "$dir/question/list.txt");
	while (<QUESTIONLIST>) {
		s/^\s+//;
		s/\s+$//;
		if (/^#/ || ($_ eq "")) {
			next;
		}
		($qid, $file) = split('\s+');
		push(@list, $qid);
	}
	close(QUESTIONLIST);

	return @list;
}

sub question_get_file {
	my $search_qid;
	my $qid;
	my $file;

	($search_qid) = @_;

	open(QUESTIONLIST, "$dir/question/list.txt");
	while (<QUESTIONLIST>) {
		s/^\s+//;
		s/\s+$//;
		if (/^#/ || ($_ eq "")) {
			next;
		}
		($qid, $file) = split('\s+');
		if ($search_qid ne $qid) {
			next;
		}
		close(QUESTIONLIST);
		return $file;
	}
	close(QUESTIONLIST);

	return "";
}

sub question_get_genre {
	my $qid;
	my $genre;
	my $score;
	my $title;
	my $question;
	my @answers;

	($qid) = @_;
	($genre, $score, $title, $question, @answers) = question_readfile($qid);
	return $genre;
}

sub question_get_score {
	my $qid;
	my $genre;
	my $score;
	my $title;
	my $question;
	my @answers;

	($qid) = @_;
	($genre, $score, $title, $question, @answers) = question_readfile($qid);
	return $score;
}

sub question_get_title {
	my $qid;
	my $genre;
	my $score;
	my $title;
	my $question;
	my @answers;

	($qid) = @_;
	($genre, $score, $title, $question, @answers) = question_readfile($qid);
	return $title;
}

sub question_readfile {
	my $qid;
	my $file;
	my $key;
	my $value;
	my $in_question;
	my $in_comment;
	my $genre;
	my $score;
	my $title;
	my $question;
	my @answers;

	($qid) = @_;

	$file = question_get_file($qid);
	if ($file eq "") {
		return ();
	}

	$in_question = 0;
	$in_comment = 0;

	open(FILE, "$dir/question/$file");
	while (<FILE>) {
		if ($in_question) {
			if (/^\s*QUESTION_END/) {
				$in_question = 0;
			} else {
				$question .= "$_\n";
			}
			next;
		}
		if ($in_comment) {
			if (/^\s*COMMENT_END/) {
				$in_comment = 0;
			}
			next;
		}

		s/^\s+//;
		s/\s+$//;

		if (/^#/ || ($_ eq "")) {
			next;
		}

		($key, $value) = /(\w+):\s+(.+)/;

		if (/^GENRE/) {
			$genre = $value;
		}
		if (/^SCORE/) {
			$score = $value;
		}
		if (/^TITLE/) {
			$title = $value;
		}
		if (/^ANSWER/) {
			push(@answers, $value);
		}
		if (/^QUESTION/) {
			$in_question = 1;
		}
		if (/^COMMENT/) {
			$in_comment = 1;
		}
	}
	close(FILE);

	return ($genre, $score, $title, $question, @answers);
}

sub question_is_clear {
	my $uid;
	my $qid;
	($uid, $qid) = @_;

	if (-f "$dir/result/$uid/${qid}.txt") {
		return 1;
	}
	return 0;
}

sub clear_get_time {
	my $uid;
	my $qid;
	($uid, $qid) = @_;

	if (!question_is_clear($uid, $qid)) {
		return 0;
	}

	open(RESULTFILE, "$dir/result/$uid/${qid}.txt");
	$t = <RESULTFILE>;
	close(RESULTFILE);

	return $t + 0;
}

sub clear_list {
	my @qid_list;
	my $uid;
	my $qid;
	my @clear_list;

	($uid) = @_;

	@qid_list = question_list();
	@clear_list = ();

	foreach $qid (@qid_list) {
		if (question_is_clear($uid, $qid)) {
			push(@clear_list, $qid);
		}
	}

	return @clear_list;
}

sub user_total_score {
	my @qid_list;
	my $uid;
	my $qid;
	my $t;
	my $total_score = 0;

	($uid, $t) = @_;

	@qid_list = clear_list($uid);

	foreach $qid (@qid_list) {
		if (defined($t) && ($t > 0)) {
			if (clear_get_time($uid, $qid) > $t) {
				next;
			}
		}
		$total_score += question_get_score($qid);
	}

	return $total_score;
}
